import sys


def test_import_check():
    # NOTE These imports should *exactly* match any hordelib imports in build_helper.py
    # If this test fails, include the imports missing in `release.yaml`.
    # and update [testenv:test-build-helper] in `tox.ini`.

    from hordelib import install_comfy
    from hordelib.consts import COMFYUI_VERSION



# This tests running hordelib standalone, as an external caller would use it.
# Call with: python -m test.run_img2img
# You need all the deps in whatever environment you are running this.
import os

import hordelib


def main():
    hordelib.initialise()

    from PIL import Image

    from hordelib.horde import HordeLib
    from hordelib.shared_model_manager import SharedModelManager

    generate = HordeLib()
    SharedModelManager.loadModelManagers(compvis=True)
    SharedModelManager.manager.load("Deliberate")

    data = {
        "sampler_name": "k_dpmpp_2m",
        "cfg_scale": 7.5,
        "denoising_strength": 0.4,
        "seed": 250636385744582,
        "height": 512,
        "width": 512,
        "karras": False,
        "tiling": False,
        "hires_fix": False,
        "clip_skip": 1,
        "control_type": None,
        "image_is_control": False,
        "return_control_map": False,
        "prompt": "a dinosaur",
        "ddim_steps": 25,
        "n_iter": 1,
        "model": "Deliberate",
        "source_image": Image.open("images/test_db0.jpg"),
    }
    pil_image = generate.basic_inference(data)
    pil_image.save("images/run_img2img.webp", quality=90)


if __name__ == "__main__":
    main()



# consts.py
import os
from enum import Enum, auto

from hordelib.config_path import get_hordelib_path

COMFYUI_VERSION = "84ea21c815d426000c233e0c7b8c542764335cc8"
"""The exact version of ComfyUI version to load."""

REMOTE_PROXY = ""

REMOTE_MODEL_DB = "https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/"
"""The default base endpoint where to find model databases. See MODEL_DB_NAMES for valid database names."""

RELEASE_VERSION = os.path.exists(os.path.join(get_hordelib_path(), "_version.py"))
"""A flag for if this is a pypi release or a git dev mode"""


class HordeSupportedBackends(Enum):
    ComfyUI = auto()


# Models Excluded from hordelib (for now) # FIXME
# this could easily be a json file on the AI-Horde-image-model-reference repo
EXCLUDED_MODEL_NAMES = ["pix2pix", "anything_v4_inpainting", "dreamlike_diffusion_inpainting"]


class MODEL_CATEGORY_NAMES(str, Enum):
    """Look up str enum for the categories of models (compvis, controlnet, clip, etc...)."""

    blip = "blip"
    clip = "clip"
    codeformer = "codeformer"
    compvis = "compvis"
    controlnet = "controlnet"
    # diffusers = "diffusers"
    esrgan = "esrgan"
    gfpgan = "gfpgan"
    safety_checker = "safety_checker"
    lora = "lora"


# Default model managers to load
DEFAULT_MODEL_MANAGERS = {
    MODEL_CATEGORY_NAMES.blip: True,
    MODEL_CATEGORY_NAMES.clip: True,
    MODEL_CATEGORY_NAMES.codeformer: True,
    MODEL_CATEGORY_NAMES.compvis: True,
    MODEL_CATEGORY_NAMES.controlnet: True,
    # MODEL_CATEGORY_NAMES.diffusers: True,
    MODEL_CATEGORY_NAMES.esrgan: True,
    MODEL_CATEGORY_NAMES.gfpgan: True,
    MODEL_CATEGORY_NAMES.safety_checker: True,
    MODEL_CATEGORY_NAMES.lora: True,
}
"""The default model managers to load."""  # XXX Clarify

MODEL_DB_NAMES = {
    MODEL_CATEGORY_NAMES.blip: MODEL_CATEGORY_NAMES.blip,
    MODEL_CATEGORY_NAMES.clip: MODEL_CATEGORY_NAMES.clip,
    MODEL_CATEGORY_NAMES.codeformer: MODEL_CATEGORY_NAMES.codeformer,
    MODEL_CATEGORY_NAMES.compvis: "stable_diffusion",
    MODEL_CATEGORY_NAMES.controlnet: MODEL_CATEGORY_NAMES.controlnet,
    # MODEL_CATEGORY_NAMES.diffusers: MODEL_CATEGORY_NAMES.diffusers,
    MODEL_CATEGORY_NAMES.esrgan: MODEL_CATEGORY_NAMES.esrgan,
    MODEL_CATEGORY_NAMES.gfpgan: MODEL_CATEGORY_NAMES.gfpgan,
    MODEL_CATEGORY_NAMES.safety_checker: MODEL_CATEGORY_NAMES.safety_checker,
    MODEL_CATEGORY_NAMES.lora: MODEL_CATEGORY_NAMES.lora,
}
"""The name of the json file (without the extension) of the corresponding model database."""

MODEL_FOLDER_NAMES = {
    MODEL_CATEGORY_NAMES.blip: MODEL_CATEGORY_NAMES.blip,
    MODEL_CATEGORY_NAMES.clip: MODEL_CATEGORY_NAMES.clip,
    MODEL_CATEGORY_NAMES.codeformer: MODEL_CATEGORY_NAMES.codeformer,
    MODEL_CATEGORY_NAMES.compvis: "compvis",
    MODEL_CATEGORY_NAMES.controlnet: MODEL_CATEGORY_NAMES.controlnet,
    # MODEL_CATEGORY_NAMES.diffusers: MODEL_CATEGORY_NAMES.diffusers,
    MODEL_CATEGORY_NAMES.esrgan: MODEL_CATEGORY_NAMES.esrgan,
    MODEL_CATEGORY_NAMES.gfpgan: MODEL_CATEGORY_NAMES.gfpgan,
    MODEL_CATEGORY_NAMES.safety_checker: MODEL_CATEGORY_NAMES.safety_checker,
    MODEL_CATEGORY_NAMES.lora: MODEL_CATEGORY_NAMES.lora,
}
"""The folder name on disk where the models are stored in AIWORKER_CACHE_HOME."""



import contextlib
import os

from pynvml.smi import nvidia_smi


class GPUInfo:
    def __init__(self):
        self.avg_load = []
        self.avg_temp = []
        self.avg_power = []
        # Average period in samples, default 10 samples per second, period 5 minutes
        self.samples_per_second = 10
        # Look out for device env var hack
        self.device = int(os.getenv("CUDA_VISIBLE_DEVICES", 0))

    # Return a value from the given dictionary supporting dot notation
    def get(self, data, key, default=""):
        # Handle nested structures
        path = key.split(".")

        if len(path) == 1:
            # Simple case
            return data[key] if key in data else default
        # Nested case
        walkdata = data
        for element in path:
            if element in walkdata:
                walkdata = walkdata[element]
            else:
                walkdata = ""
                break
        return walkdata

    def _get_gpu_data(self):
        with contextlib.suppress(Exception):
            nvsmi = nvidia_smi.getInstance()
            data = nvsmi.DeviceQuery()
            return data.get("gpu", [None])[self.device]

    def _mem(self, raw):
        unit = "GB"
        mem = raw / 1024
        if mem < 1:
            unit = "MB"
            raw *= 1024
        return f"{round(mem)} {unit}"

    def get_total_vram_mb(self):
        """Get total VRAM in MB as an integer, or 0"""
        value = 0
        data = self._get_gpu_data()
        if data:
            value = self.get(data, "fb_memory_usage.total", "0")
            try:
                value = int(value)
            except ValueError:
                value = 0
        return value

    def get_free_vram_mb(self):
        """Get free VRAM in MB as an integer, or 0"""
        value = 0
        data = self._get_gpu_data()
        if data:
            value = self.get(data, "fb_memory_usage.free", "0")
            try:
                value = int(value)
            except ValueError:
                value = 0
        return value

    def get_info(self):
        data = self._get_gpu_data()
        if not data:
            return None

        # Calculate averages
        try:
            gpu_util = int(self.get(data, "utilization.gpu_util", 0))
        except ValueError:
            gpu_util = 0

        try:
            gpu_temp = int(self.get(data, "temperature.gpu_temp", 0))
        except ValueError:
            gpu_temp = 0

        try:
            gpu_power = int(self.get(data, "power_readings.power_draw", 0))
        except ValueError:
            gpu_power = 0

        self.avg_load.append(gpu_util)
        self.avg_temp.append(gpu_temp)
        self.avg_power.append(gpu_power)
        self.avg_load = self.avg_load[-(self.samples_per_second * 60 * 5) :]
        self.avg_power = self.avg_power[-(self.samples_per_second * 60 * 5) :]
        self.avg_temp = self.avg_temp[-(self.samples_per_second * 60 * 5) :]
        avg_load = int(sum(self.avg_load) / len(self.avg_load))
        avg_power = int(sum(self.avg_power) / len(self.avg_power))
        avg_temp = int(sum(self.avg_temp) / len(self.avg_temp))

        return {
            "product": self.get(data, "product_name", "unknown"),
            "pci_gen": self.get(data, "pci.pci_gpu_link_info.pcie_gen.current_link_gen", "?"),
            "pci_width": self.get(data, "pci.pci_gpu_link_info.link_widths.current_link_width", "?"),
            "fan_speed": f"{self.get(data, 'fan_speed')}{self.get(data, 'fan_speed_unit')}",
            "vram_total": f"{self._mem(self.get(data, 'fb_memory_usage.total', '0'))}",
            "vram_used": f"{self._mem(self.get(data, 'fb_memory_usage.used', '0'))}",
            "vram_free": f"{self._mem(self.get(data, 'fb_memory_usage.free', '0'))}",
            "load": f"{gpu_util}{self.get(data, 'utilization.unit')}",
            "temp": f"{gpu_temp}{self.get(data, 'temperature.unit')}",
            "power": f"{gpu_power}{self.get(data, 'power_readings.unit')}",
            "avg_load": f"{avg_load}{self.get(data, 'utilization.unit')}",
            "avg_temp": f"{avg_temp}{self.get(data, 'temperature.unit')}",
            "avg_power": f"{avg_power}{self.get(data, 'power_readings.unit')}",
        }



import pytest

import hordelib
import hordelib.consts as consts


class TestCompvis:
    _initialised = False

    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        if not self._initialised:
            hordelib.initialise()
        from hordelib.model_manager.compvis import CompVisModelManager

        self.compvis_model_manager = CompVisModelManager()
        assert self.compvis_model_manager is not None
        yield
        del self.compvis_model_manager

    def test_compvis_load_defaults(self):
        success = self.compvis_model_manager.load("Deliberate")
        assert success



# This tests running hordelib standalone, as an external caller would use it.
# Call with: python -m test.run_clip
# You need all the deps in whatever environment you are running this.
import os

import hordelib


def main():
    hordelib.initialise()

    import json

    from loguru import logger
    from PIL import Image

    from hordelib.blip.caption import Caption
    from hordelib.clip.interrogate import Interrogator
    from hordelib.horde import HordeLib
    from hordelib.shared_model_manager import SharedModelManager

    HordeLib()
    SharedModelManager.loadModelManagers(clip=True, blip=True)
    SharedModelManager.manager.load("ViT-L/14")
    SharedModelManager.manager.load("BLIP_Large")

    # Run CLIP
    model_info = SharedModelManager.manager.loaded_models["ViT-L/14"]
    interrogator = Interrogator(model_info)
    ranking_result = interrogator(
        image=Image.open("images/test_inpaint_original.png"),
        rank=True,
    )
    logger.warning(json.dumps(ranking_result, indent=4))

    # Run BLIP
    model = SharedModelManager.manager.loaded_models["BLIP_Large"]
    caption_class = Caption(model)
    caption = caption_class(
        image=Image.open("images/test_inpaint_original.png"),
        sample=True,
        num_beams=7,
        min_length=20,
        max_length=50,
        top_p=0.9,
        repetition_penalty=1.4,
    )
    logger.warning(json.dumps(caption, indent=4))


if __name__ == "__main__":
    main()



import torch

from nodes import MAX_RESOLUTION

class LatentCompositeMasked:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "destination": ("LATENT",),
                "source": ("LATENT",),
                "x": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 8}),
                "y": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 8}),
            },
            "optional": {
                "mask": ("MASK",),
            }
        }
    RETURN_TYPES = ("LATENT",)
    FUNCTION = "composite"

    CATEGORY = "latent"

    def composite(self, destination, source, x, y, mask = None):
        output = destination.copy()
        destination = destination["samples"].clone()
        source = source["samples"]

        x = max(-source.shape[3] * 8, min(x, destination.shape[3] * 8))
        y = max(-source.shape[2] * 8, min(y, destination.shape[2] * 8))

        left, top = (x // 8, y // 8)
        right, bottom = (left + source.shape[3], top + source.shape[2],)


        if mask is None:
            mask = torch.ones_like(source)
        else:
            mask = mask.clone()
            mask = torch.nn.functional.interpolate(mask[None, None], size=(source.shape[2], source.shape[3]), mode="bilinear")
            mask = mask.repeat((source.shape[0], source.shape[1], 1, 1))

        # calculate the bounds of the source that will be overlapping the destination
        # this prevents the source trying to overwrite latent pixels that are out of bounds
        # of the destination
        visible_width, visible_height = (destination.shape[3] - left + min(0, x), destination.shape[2] - top + min(0, y),)

        mask = mask[:, :, :visible_height, :visible_width]
        inverse_mask = torch.ones_like(mask) - mask

        source_portion = mask * source[:, :, :visible_height, :visible_width]
        destination_portion = inverse_mask  * destination[:, :, top:bottom, left:right]

        destination[:, :, top:bottom, left:right] = source_portion + destination_portion

        output["samples"] = destination

        return (output,)

class MaskToImage:
    @classmethod
    def INPUT_TYPES(s):
        return {
                "required": {
                    "mask": ("MASK",),
                }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "mask_to_image"

    def mask_to_image(self, mask):
        result = mask.reshape((-1, 1, mask.shape[-2], mask.shape[-1])).movedim(1, -1).expand(-1, -1, -1, 3)
        return (result,)

class ImageToMask:
    @classmethod
    def INPUT_TYPES(s):
        return {
                "required": {
                    "image": ("IMAGE",),
                    "channel": (["red", "green", "blue"],),
                }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)
    FUNCTION = "image_to_mask"

    def image_to_mask(self, image, channel):
        channels = ["red", "green", "blue"]
        mask = image[0, :, :, channels.index(channel)]
        return (mask,)

class SolidMask:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "value": ("FLOAT", {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01}),
                "width": ("INT", {"default": 512, "min": 1, "max": MAX_RESOLUTION, "step": 1}),
                "height": ("INT", {"default": 512, "min": 1, "max": MAX_RESOLUTION, "step": 1}),
            }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)

    FUNCTION = "solid"

    def solid(self, value, width, height):
        out = torch.full((height, width), value, dtype=torch.float32, device="cpu")
        return (out,)

class InvertMask:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "mask": ("MASK",),
            }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)

    FUNCTION = "invert"

    def invert(self, mask):
        out = 1.0 - mask
        return (out,)

class CropMask:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "mask": ("MASK",),
                "x": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "y": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "width": ("INT", {"default": 512, "min": 1, "max": MAX_RESOLUTION, "step": 1}),
                "height": ("INT", {"default": 512, "min": 1, "max": MAX_RESOLUTION, "step": 1}),
            }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)

    FUNCTION = "crop"

    def crop(self, mask, x, y, width, height):
        out = mask[y:y + height, x:x + width]
        return (out,)

class MaskComposite:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "destination": ("MASK",),
                "source": ("MASK",),
                "x": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "y": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "operation": (["multiply", "add", "subtract"],),
            }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)

    FUNCTION = "combine"

    def combine(self, destination, source, x, y, operation):
        output = destination.clone()

        left, top = (x, y,)
        right, bottom = (min(left + source.shape[1], destination.shape[1]), min(top + source.shape[0], destination.shape[0]))
        visible_width, visible_height = (right - left, bottom - top,)

        source_portion = source[:visible_height, :visible_width]
        destination_portion = destination[top:bottom, left:right]

        if operation == "multiply":
            output[top:bottom, left:right] = destination_portion * source_portion
        elif operation == "add":
            output[top:bottom, left:right] = destination_portion + source_portion
        elif operation == "subtract":
            output[top:bottom, left:right] = destination_portion - source_portion

        output = torch.clamp(output, 0.0, 1.0)

        return (output,)

class FeatherMask:
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "mask": ("MASK",),
                "left": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "top": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "right": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
                "bottom": ("INT", {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 1}),
            }
        }

    CATEGORY = "mask"

    RETURN_TYPES = ("MASK",)

    FUNCTION = "feather"

    def feather(self, mask, left, top, right, bottom):
        output = mask.clone()

        left = min(left, output.shape[1])
        right = min(right, output.shape[1])
        top = min(top, output.shape[0])
        bottom = min(bottom, output.shape[0])

        for x in range(left):
            feather_rate = (x + 1.0) / left
            output[:, x] *= feather_rate

        for x in range(right):
            feather_rate = (x + 1) / right
            output[:, -x] *= feather_rate

        for y in range(top):
            feather_rate = (y + 1) / top
            output[y, :] *= feather_rate

        for y in range(bottom):
            feather_rate = (y + 1) / bottom
            output[-y, :] *= feather_rate

        return (output,)



NODE_CLASS_MAPPINGS = {
    "LatentCompositeMasked": LatentCompositeMasked,
    "MaskToImage": MaskToImage,
    "ImageToMask": ImageToMask,
    "SolidMask": SolidMask,
    "InvertMask": InvertMask,
    "CropMask": CropMask,
    "MaskComposite": MaskComposite,
    "FeatherMask": FeatherMask,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "ImageToMask": "Convert Image to Mask",
    "MaskToImage": "Convert Mask to Image",
}



# horde.py
# Main interface for the horde to this library.
import glob
import json
import os
import random
import sys
import time
from copy import deepcopy

from loguru import logger
from PIL import Image

from hordelib.comfy_horde import Comfy_Horde
from hordelib.shared_model_manager import SharedModelManager
from hordelib.utils.dynamicprompt import DynamicPromptParser
from hordelib.utils.image_utils import ImageUtils
from hordelib.utils.sanitizer import Sanitizer


class HordeLib:
    _instance = None
    _initialised = False

    # Horde to comfy sampler mapping
    SAMPLERS_MAP = {
        "k_euler": "euler",
        "k_euler_a": "euler_ancestral",
        "k_heun": "heun",
        "k_dpm_2": "dpm_2",
        "k_dpm_2_a": "dpm_2_ancestral",
        "k_lms": "lms",
        "k_dpm_fast": "dpm_fast",
        "k_dpm_adaptive": "dpm_adaptive",
        "k_dpmpp_2s_a": "dpmpp_2s_ancestral",
        "k_dpmpp_sde": "dpmpp_sde",
        "k_dpmpp_2m": "dpmpp_2m",
        "ddim": "ddim",
        "uni_pc": "uni_pc",
        "uni_pc_bh2": "uni_pc_bh2",
        "plms": "euler",
    }

    # Horde names on the left, our node names on the right
    # We use this to dynamically route the image through the
    # right node by reconnect inputs.
    CONTROLNET_IMAGE_PREPROCESSOR_MAP = {
        "canny": "canny",
        "hed": "hed",
        "depth": "depth",
        "normal": "normal",
        "openpose": "openpose",
        "seg": "seg",
        "scribble": "scribble",
        "fakescribbles": "fakescribble",
        "hough": "mlsd",  # horde backward compatibility
        "mlsd": "mlsd",
        # "<unused>": "MiDaS-DepthMapPreprocessor",
        # "<unused>": "MediaPipe-HandPosePreprocessor",
        # "<unused>": "MediaPipe-FaceMeshPreprocessor",
        # "<unused>": "BinaryPreprocessor",
        # "<unused>": "ColorPreprocessor",
        # "<unused>": "PiDiNetPreprocessor",
    }

    SOURCE_IMAGE_PROCESSING_OPTIONS = ["img2img", "inpainting", "outpainting"]

    SCHEDULERS = ["normal", "karras", "simple", "ddim_uniform"]

    # Describe a valid payload, it's types and bounds. All incoming payload data is validated against,
    # and normalised to, this schema.
    PAYLOAD_SCHEMA = {
        "sampler_name": {"datatype": str, "values": list(SAMPLERS_MAP.keys()), "default": "k_euler"},
        "cfg_scale": {"datatype": float, "min": 1, "max": 100, "default": 8.0},
        "denoising_strength": {"datatype": float, "min": 0.01, "max": 1.0, "default": 1.0},
        "control_strength": {"datatype": float, "min": 0.01, "max": 1.0, "default": 1.0},
        "seed": {"datatype": int, "default": random.randint(0, sys.maxsize)},
        "width": {"datatype": int, "min": 64, "max": 8192, "default": 512, "divisible": 64},
        "height": {"datatype": int, "min": 64, "max": 8192, "default": 512, "divisible": 64},
        "hires_fix": {"datatype": bool, "default": False},
        "clip_skip": {"datatype": int, "min": 1, "max": 20, "default": 1},
        "control_type": {"datatype": str, "values": list(CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys()), "default": None},
        "image_is_control": {"datatype": bool, "default": False},
        "return_control_map": {"datatype": bool, "default": False},
        "prompt": {"datatype": str, "default": ""},
        "negative_prompt": {"datatype": str, "default": ""},
        "loras": {"datatype": list, "default": []},
        "ddim_steps": {"datatype": int, "min": 1, "max": 500, "default": 30},
        "n_iter": {"datatype": int, "min": 1, "max": 100, "default": 1},
        "model": {"datatype": str, "default": "stable_diffusion"},
        "source_mask": {"datatype": Image.Image, "default": None},
        "source_image": {"datatype": Image.Image, "default": None},
        "source_processing": {"datatype": str, "values": SOURCE_IMAGE_PROCESSING_OPTIONS, "default": None},
        "hires_fix_denoising_strength": {"datatype": float, "min": 0.0, "max": 1.0, "default": 0.65},
        "scheduler": {"datatype": str, "values": SCHEDULERS, "default": "normal"},
    }

    LORA_SCHEMA = {
        "name": {"datatype": str, "default": ""},
        "model": {"datatype": float, "min": 0.0, "max": 1.0, "default": 1.0},
        "clip": {"datatype": float, "min": 0.0, "max": 1.0, "default": 1.0},
        "inject_trigger": {"datatype": str},
    }

    # pipeline parameter <- hordelib payload parameter mapping
    PAYLOAD_TO_PIPELINE_PARAMETER_MAPPING = {
        "sampler.sampler_name": "sampler_name",
        "sampler.cfg": "cfg_scale",
        "sampler.denoise": "denoising_strength",
        "sampler.seed": "seed",
        "empty_latent_image.height": "height",
        "empty_latent_image.width": "width",
        "sampler.scheduler": "scheduler",
        "clip_skip.stop_at_clip_layer": "clip_skip",
        "prompt.text": "prompt",
        "negative_prompt.text": "negative_prompt",
        "sampler.steps": "ddim_steps",
        "empty_latent_image.batch_size": "n_iter",
        "model_loader.model_name": "model",
        "image_loader.image": "source_image",
        "loras": "loras",
        "upscale_sampler.denoise": "hires_fix_denoising_strength",
        "upscale_sampler.seed": "seed",
        "upscale_sampler.cfg": "cfg_scale",
        "upscale_sampler.steps": "ddim_steps",
        "upscale_sampler.sampler_name": "sampler_name",
        "controlnet_apply.strength": "control_strength",
        "controlnet_model_loader.control_net_name": "control_type",
    }

    # We are a singleton
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # We initialise only ever once (in the lifetime of the singleton)
    def __init__(self):
        if not self._initialised:
            self.generator = Comfy_Horde()
            self.__class__._initialised = True

    def _json_hack(self, obj):
        # Helper to serialise json which contains non-serialisable types
        if hasattr(obj, "__class__"):
            return f"{obj.__class__.__name__} instance"
        return f"Object of type {type(obj).__name__}"

    def dump_json(self, adict):
        logger.warning(json.dumps(adict, indent=4, default=self._json_hack))

    def _validate(self, value, datatype, min=None, max=None, default=None, values=None, divisible=None):
        """Check the given value against the given constraints. Return the fixed value."""

        # First, if we are passed no value at all, use the default
        if value is None:
            return default

        # We have a value, check the type
        if not isinstance(value, datatype):
            # try to coerce the type
            try:
                value = datatype(value)
            except (ValueError, TypeError):
                # epic fail, use the default
                return default

        # If the value must be divisible by some amount, assert that
        if divisible:
            if value % divisible != 0:
                value = ((value + (divisible - 1)) // divisible) * divisible

        # If we have a minimum, assert it
        if min and value < min:
            value = min

        # If we have a maximum, assert it
        if max and value > max:
            value = max

        # If we have a list of allowed values, make sure the value is permitted
        if values:
            if isinstance(value, str):
                if value.lower() not in values:
                    # not an allowed value, use the default
                    return default
                else:
                    value = value.lower()
            else:
                if value not in values:
                    # not an allowed value, use the default
                    return default

        return value

    def _validate_data_structure(self, data, schema_definition=PAYLOAD_SCHEMA):
        """Validate a data structure, assert parameters fall within the allowed bounds."""
        data = deepcopy(data)

        # Remove anything from the payload that isn't in our schema
        for key in data.copy().keys():
            if key.lower() not in schema_definition.keys():
                del data[key]

        # Build a valid schema payload, create attributes that don't exist using default values
        for key, schema in schema_definition.items():
            data[key.lower()] = self._validate(data.get(key.lower()), **schema)

        # Do the same for loras, if we have loras in this data structure
        if data.get("loras"):
            for i, lora in enumerate(data.get("loras")):
                data["loras"][i] = self._validate_data_structure(lora, HordeLib.LORA_SCHEMA)
            # Remove invalid loras
            data["loras"] = [x for x in data["loras"] if x.get("name")]

        return data

    def _apply_aihorde_compatibility_hacks(self, payload):
        """For use by the AI Horde worker we require various counterintuitive hacks to the payload data.

        We encapsulate all of this implicit witchcraft in one function, here.
        """
        payload = deepcopy(payload)

        # Rather than specify a scheduler, only karras or not karras is specified
        if payload.get("karras", False):
            payload["scheduler"] = "karras"
        else:
            payload["scheduler"] = "normal"

        # Negative and positive prompts are merged together
        if payload.get("prompt"):
            split_prompts = [x.strip() for x in payload.get("prompt").split("###")][:2]
            if len(split_prompts) == 2:
                payload["prompt"] = split_prompts[0]
                payload["negative_prompt"] = split_prompts[1]

        # Turn off hires fix if we're not generating a hires image, or if the params are just confused
        try:
            if "hires_fix" in payload and (payload["width"] <= 512 or payload["height"] <= 512):
                payload["hires_fix"] = False
        except (TypeError, KeyError):
            payload["hires_fix"] = False

        # Turn off hires fix if we're inpainting as the dimensions are from the source image
        if "hires_fix" in payload and (
            payload.get("source_processing") == "inpainting" or payload.get("source_processing") == "outpainting"
        ):
            payload["hires_fix"] = False

        # Use denoising strength for both samplers if no second denoiser specified
        # but not for txt2img where denoising will always generally be 1.0
        if payload.get("hires_fix"):
            if payload.get("source_processing") and payload.get("source_processing") != "txt2img":
                if not payload.get("hires_fix_denoising_strength"):
                    payload["hires_fix_denoising_strength"] = payload.get("denoising_strength")

        # Remap "denoising" to "controlnet strength", historical hack
        if payload.get("control_type"):
            if payload.get("denoising_strength"):
                if not payload.get("control_strength"):
                    payload["control_strength"] = payload["denoising_strength"]
                    del payload["denoising_strength"]
                else:
                    del payload["denoising_strength"]

        return payload

    def _final_pipeline_adjustments(self, payload, pipeline_data):

        payload = deepcopy(payload)

        # Process dynamic prompts
        if new_prompt := DynamicPromptParser(payload["seed"]).parse(payload.get("prompt", "")):
            payload["prompt"] = new_prompt

        # Clip skip inversion, comfy uses -1, -2, etc
        if payload.get("clip_skip", 0) > 0:
            payload["clip_skip"] = -payload["clip_skip"]

        # Remap sampler name, horde to comfy
        payload["sampler_name"] = HordeLib.SAMPLERS_MAP.get(payload["sampler_name"])

        # Remap controlnet models, horde to comfy
        if payload.get("control_type"):
            payload["control_type"] = HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.get(payload.get("control_type"))

        # Setup controlnet if required
        # For LORAs we completely build the LORA section of the pipeline dynamically, as we have
        # to handle n LORA models which form chained nodes in the pipeline.
        # Note that we build this between several nodes, the model_loader, clip_skip and the sampler,
        # plus the upscale sampler (used in hires fix) if there is one
        if payload.get("loras") and SharedModelManager.manager.lora:

            # Remove any requested LORAs that we don't have
            valid_loras = []
            for lora in payload.get("loras"):
                # Determine the actual lora filename
                if SharedModelManager.manager.lora.is_local_model(str(lora["name"])):
                    # We store the actual lora name to search for the trigger
                    lora_name = SharedModelManager.manager.lora.get_lora_name(str(lora["name"]))
                    if lora_name:
                        logger.debug(f"Found valid lora {lora_name}")
                        trigger_inject = lora.get("inject_trigger")
                        trigger = None
                        if trigger_inject == "any":
                            triggers = SharedModelManager.manager.lora.get_lora_triggers(lora_name)
                            if triggers:
                                trigger = triggers[0]
                        elif trigger_inject is not None:
                            trigger = SharedModelManager.manager.lora.find_lora_trigger(lora_name, trigger_inject)
                        if trigger:
                            # We inject at the start, to avoid throwing it in a negative prompt
                            payload["prompt"] = f'{trigger}, {payload["prompt"]}'
                        # the fixed up and validated filename (Comfy expect the "name" key to be the filename)
                        lora["name"] = SharedModelManager.manager.lora.get_lora_filename(lora_name)
                        valid_loras.append(lora)
            payload["loras"] = valid_loras
            for lora_index, lora in enumerate(payload.get("loras")):

                # Inject a lora node (first lora)
                if lora_index == 0:
                    pipeline_data[f"lora_{lora_index}"] = {
                        "inputs": {
                            "model": ["model_loader", 0],
                            "clip": ["model_loader", 1],
                            "lora_name": lora["name"],
                            "strength_model": lora["model"],
                            "strength_clip": lora["clip"],
                            "model_manager": SharedModelManager,
                        },
                        "class_type": "HordeLoraLoader",
                    }
                else:
                    # Subsequent chained loras
                    pipeline_data[f"lora_{lora_index}"] = {
                        "inputs": {
                            "model": [f"lora_{lora_index-1}", 0],
                            "clip": [f"lora_{lora_index-1}", 1],
                            "lora_name": lora["name"],
                            "strength_model": lora["model"],
                            "strength_clip": lora["clip"],
                            "model_manager": SharedModelManager,
                        },
                        "class_type": "HordeLoraLoader",
                    }

            for lora_index, lora in enumerate(payload.get("loras")):

                # The first LORA always connects to the model loader
                if lora_index == 0:
                    self.generator.reconnect_input(pipeline_data, "lora_0.model", "model_loader")
                    self.generator.reconnect_input(pipeline_data, "lora_0.clip", "model_loader")
                else:
                    # Other loras connect to the previous lora
                    self.generator.reconnect_input(
                        pipeline_data,
                        f"lora_{lora_index}.model",
                        f"lora_{lora_index-1}.model",
                    )
                    self.generator.reconnect_input(
                        pipeline_data,
                        f"lora_{lora_index}.clip",
                        f"lora_{lora_index-1}.clip",
                    )

                # The last LORA always connects to the sampler and clip text encoders (via the clip_skip)
                if lora_index == len(payload.get("loras")) - 1:
                    self.generator.reconnect_input(pipeline_data, "sampler.model", f"lora_{lora_index}")
                    self.generator.reconnect_input(pipeline_data, "upscale_sampler.model", f"lora_{lora_index}")
                    self.generator.reconnect_input(pipeline_data, "clip_skip.clip", f"lora_{lora_index}")

        # Translate the payload parameters into pipeline parameters
        pipeline_params = {}
        for newkey, key in HordeLib.PAYLOAD_TO_PIPELINE_PARAMETER_MAPPING.items():
            if key in payload:
                pipeline_params[newkey] = payload.get(key)
            else:
                logger.error(f"Parameter {key} not found")

        # Inject our model manager
        pipeline_params["model_loader.model_manager"] = SharedModelManager

        # For hires fix, change the image sizes as we create an intermediate image first
        if payload.get("hires_fix", False):
            width = pipeline_params.get("empty_latent_image.width", 0)
            height = pipeline_params.get("empty_latent_image.height", 0)
            if width > 512 and height > 512:
                newwidth, newheight = ImageUtils.calculate_source_image_size(width, height)
                pipeline_params["latent_upscale.width"] = width
                pipeline_params["latent_upscale.height"] = height
                pipeline_params["empty_latent_image.width"] = newwidth
                pipeline_params["empty_latent_image.height"] = newheight

        if payload.get("control_type"):
            # Inject control net model manager
            pipeline_params["controlnet_model_loader.model_manager"] = SharedModelManager

            # Dynamically reconnect nodes in the pipeline to connect the correct pre-processor node
            if payload.get("return_control_map"):
                # Connect annotator to output image directly if we need to return the control map
                self.generator.reconnect_input(
                    pipeline_data,
                    "output_image.images",
                    payload["control_type"],
                )
            else:
                # Connect annotator to controlnet apply node
                self.generator.reconnect_input(
                    pipeline_data,
                    "controlnet_apply.image",
                    payload["control_type"],
                )

        # If we have a source image, use that rather than latent noise (i.e. img2img)
        # We do this by reconnecting the nodes in the pipeline to make the input to the vae encoder
        # the source image instead of the latent noise generator
        if pipeline_params.get("image_loader.image"):
            self.generator.reconnect_input(pipeline_data, "sampler.latent_image", "vae_encode")

        return pipeline_params

    def _get_appropriate_pipeline(self, params):
        # Determine the correct pipeline based on the parameters we have
        #
        # The pipelines are:
        #
        #     stable_diffusion
        #       stable_diffusion_hires_fix
        #     stable_diffusion_img2img_mask
        #     stable_diffusion_paint
        #     controlnet
        #       controlnet_hires_fix
        #       controlnet_annotator
        #     image_facefix
        #     image_upscale

        # controlnet, controlnet_hires_fix controlnet_annotator
        if params.get("control_type"):
            if params.get("return_control_map", False):
                return "controlnet_annotator"
            else:
                if params.get("hires_fix"):
                    return "controlnet_hires_fix"
                else:
                    return "controlnet"

        # stable_diffusion_paint, stable_diffusion_img2img_mask
        if params.get("source_processing") == "img2img":
            has_mask = params.get("source_mask") or (
                params.get("source_image") and len(params.get("source_image", "").getbands()) == 4
            )
            if has_mask:
                return "stable_diffusion_img2img_mask"
        elif params.get("source_processing") == "inpainting":
            return "stable_diffusion_paint"
        elif params.get("source_processing") == "outpainting":
            return "stable_diffusion_paint"

        # stable_diffusion and stable_diffusion_hires_fix
        if params.get("hires_fix", False):
            return "stable_diffusion_hires_fix"
        else:
            return "stable_diffusion"  # also includes img2img mode

    def _process_results(self, images, rawpng):
        if images is None:
            return None  # XXX Log error and/or raise Exception here
        # Return image(s) or raw PNG bytestream
        if not rawpng:
            results = [Image.open(x["imagedata"]) for x in images]
        else:
            results = [x["imagedata"] for x in images]
        if len(results) == 1:
            return results[0]
        else:
            return results

    def lock_models(self, models):
        models = [str(x).strip() for x in models if x]
        # Try to acquire a model lock, if we can't, wait a while as some other thread
        # must have these resources locked
        while not self.generator.lock_models(models):
            time.sleep(0.1)
        logger.debug(f"Locked models {','.join(models)}")

    def unlock_models(self, models):
        models = [x.strip() for x in models if x]
        self.generator.unlock_models(models)
        logger.debug(f"Unlocked models {','.join(models)}")

    def basic_inference(self, payload, rawpng=False):
        # AIHorde hacks to payload
        payload = self._apply_aihorde_compatibility_hacks(payload)
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Resize the source image and mask to actual final width/height requested
        ImageUtils.resize_sources_to_request(payload)
        # Determine the correct pipeline to use
        pipeline = self._get_appropriate_pipeline(payload)
        # Final adjustments to the pipeline
        pipeline_data = self.generator.get_pipeline_data(pipeline)
        payload = self._final_pipeline_adjustments(payload, pipeline_data)
        # Run the pipeline
        try:
            # Add prefix to loras to avoid name collisions with other models
            models = [f"lora-{x['name']}" for x in payload.get("loras", []) if x]
            # main model
            models.append(payload.get("model"))
            # controlnet model
            models.append(payload.get("control_type"))
            # Acquire a lock on all these models
            self.lock_models(models)
            # Call the inference pipeline
            # logger.info(payload)
            images = self.generator.run_image_pipeline(pipeline_data, payload)
        finally:
            self.unlock_models(models)
        return self._process_results(images, rawpng)

    def image_upscale(self, payload, rawpng=False) -> Image.Image | None:
        # AIHorde hacks to payload
        payload = self._apply_aihorde_compatibility_hacks(payload)
        # Remember if we were passed width and height, we wouldn't normally be passed width and height
        # because the upscale models upscale to a fixed multiple of image size. However, if we *are*
        # passed a width and height we rescale the upscale output image to this size.
        width = payload.get("width")
        height = payload.get("width")
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Final adjustments to the pipeline
        pipeline_name = "image_upscale"
        pipeline_data = self.generator.get_pipeline_data(pipeline_name)
        payload = self._final_pipeline_adjustments(payload, pipeline_data)
        # Run the pipeline
        try:
            self.lock_models([payload.get("model")])
            images = self.generator.run_image_pipeline(pipeline_data, payload)
        finally:
            self.unlock_models([payload.get("model")])
        if images is None:
            return None  # XXX Log error and/or raise Exception here
        # Allow arbitrary resizing by shrinking the image back down
        if width or height:
            return ImageUtils.shrink_image(Image.open(images[0]["imagedata"]), width, height)
        return self._process_results(images, rawpng)

    def image_facefix(self, payload, rawpng=False) -> Image.Image | None:
        # AIHorde hacks to payload
        payload = self._apply_aihorde_compatibility_hacks(payload)
        # Check payload types/values and normalise it's format
        payload = self._validate_data_structure(payload)
        # Final adjustments to the pipeline
        pipeline_name = "image_facefix"
        pipeline_data = self.generator.get_pipeline_data(pipeline_name)
        payload = self._final_pipeline_adjustments(payload, pipeline_data)
        # Run the pipeline
        try:
            self.lock_models([payload.get("model")])
            images = self.generator.run_image_pipeline(pipeline_data, payload)
        finally:
            self.unlock_models([payload.get("model")])
        return self._process_results(images, rawpng)



# initialisation.py
# Initialise hordelib.
import os
import shutil
import sys

from loguru import logger

from hordelib import install_comfy
from hordelib.config_path import get_hordelib_path, set_system_path
from hordelib.consts import (
    COMFYUI_VERSION,
    DEFAULT_MODEL_MANAGERS,
    MODEL_CATEGORY_NAMES,
    RELEASE_VERSION,
)
from hordelib.utils.logger import HordeLog

_is_initialised = False


def initialise(
    # model_managers_to_load: dict[MODEL_CATEGORY_NAMES, bool] = DEFAULT_MODEL_MANAGERS,
    *,
    setup_logging=True,
    clear_logs=False,
    debug_logging=False,
):  # XXX # TODO Do we need `model_managers_to_load`?

    global _is_initialised

    # Wipe existing logs if requested
    if clear_logs and os.path.exists("./logs"):
        shutil.rmtree("./logs")

    # Setup logging if requested
    HordeLog.initialise(setup_logging=setup_logging)
    if debug_logging:
        HordeLog.set_logger_verbosity(5)
        HordeLog.quiesce_logger(0)

    # If developer mode, don't permit some things
    if not RELEASE_VERSION and " " in get_hordelib_path():
        # Our runtime patching can't handle this
        raise Exception(
            "Do not run this project in developer mode from a path that " "contains spaces in directory names.",
        )

    # Ensure we have ComfyUI
    logger.debug("Clearing command line args in sys.argv before ComfyUI load")
    sys_arg_bkp = sys.argv.copy()
    sys.argv = sys.argv[:1]
    installer = install_comfy.Installer()
    installer.install(COMFYUI_VERSION)

    # Modify python path to include comfyui
    set_system_path()

    # Initialise model manager
    from hordelib.shared_model_manager import SharedModelManager

    SharedModelManager()

    sys.argv = sys_arg_bkp

    _is_initialised = True


def is_initialised():
    return _is_initialised



# test_horde.py
import pytest
from PIL import Image

from hordelib.horde import HordeLib
from hordelib.shared_model_manager import SharedModelManager


class TestHordeUpscaling:
    @pytest.fixture(autouse=True, scope="class")
    def setup_and_teardown(self):
        TestHordeUpscaling.horde = HordeLib()

        TestHordeUpscaling.image = Image.open("images/test_db0.jpg")
        TestHordeUpscaling.real_image = Image.open("images/test_annotator.jpg")
        (
            TestHordeUpscaling.width,
            TestHordeUpscaling.height,
        ) = TestHordeUpscaling.image.size
        yield
        del TestHordeUpscaling.horde

    @pytest.fixture(autouse=True)
    def setup_model(self, request):
        mm_type = request.node.get_closest_marker("mm_model").args[0]
        print(mm_type)
        self.default_model_manager_args = {
            mm_type: True,
        }
        SharedModelManager.loadModelManagers(**self.default_model_manager_args)
        assert SharedModelManager.manager is not None
        yield
        SharedModelManager._instance = None
        SharedModelManager.manager = None

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_RealESRGAN_x4plus(self):
        SharedModelManager.manager.load("RealESRGAN_x4plus")
        assert SharedModelManager.manager.esrgan.is_model_loaded("RealESRGAN_x4plus") is True
        data = {
            "model": "RealESRGAN_x4plus",
            "source_image": self.image,
        }
        assert self.horde is not None
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == self.width * 4
        assert height == self.height * 4
        pil_image.save("images/image_upscale_RealESRGAN_x4plus.webp", quality=90)

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_RealESRGAN_x2plus(self):
        SharedModelManager.manager.load("RealESRGAN_x2plus")
        assert SharedModelManager.manager.esrgan.is_model_loaded("RealESRGAN_x2plus") is True
        data = {
            "model": "RealESRGAN_x2plus",
            "source_image": self.image,
        }
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == self.width * 2
        assert height == self.height * 2
        pil_image.save("images/image_upscale_RealESRGAN_x2plus.webp", quality=90)

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_NMKD_Siax(self):
        SharedModelManager.manager.load("NMKD_Siax")
        assert SharedModelManager.manager.esrgan.is_model_loaded("NMKD_Siax") is True
        data = {
            "model": "NMKD_Siax",
            "source_image": self.image,
        }
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == self.width * 4
        assert height == self.height * 4
        pil_image.save("images/image_upscale_NMKD_Siax.webp", quality=90)

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_NMKD_Siax_resize(self):
        SharedModelManager.manager.load("NMKD_Siax")
        assert SharedModelManager.manager.esrgan.is_model_loaded("NMKD_Siax") is True
        data = {"model": "NMKD_Siax", "source_image": self.real_image, "width": 1280, "height": 1280}
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == 1280
        assert height == 1280
        pil_image.save("images/image_upscale_NMKD_Siax_resize.webp", quality=90)

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_RealESRGAN_x4plus_anime_6B(self):
        SharedModelManager.manager.load("RealESRGAN_x4plus_anime_6B")
        assert (
            SharedModelManager.manager.esrgan.is_model_loaded(
                "RealESRGAN_x4plus_anime_6B",
            )
            is True
        )
        data = {
            "model": "RealESRGAN_x4plus_anime_6B",
            "source_image": self.image,
        }
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == self.width * 4
        assert height == self.height * 4
        pil_image.save(
            "images/image_upscale_RealESRGAN_x4plus_anime_6B.webp",
            quality=90,
        )

    @pytest.mark.mm_model("esrgan")
    def test_image_upscale_4x_AnimeSharp(self):
        SharedModelManager.manager.load("4x_AnimeSharp")
        assert SharedModelManager.manager.esrgan.is_model_loaded("4x_AnimeSharp") is True
        data = {
            "model": "4x_AnimeSharp",
            "source_image": self.image,
        }
        pil_image = self.horde.image_upscale(data)
        assert pil_image is not None
        width, height = pil_image.size
        assert width == self.width * 4
        assert height == self.height * 4
        pil_image.save("images/image_upscale_4x_AnimeSharp.webp", quality=90)

    @pytest.mark.mm_model("codeformer")
    def test_image_facefix_codeformers(self):
        SharedModelManager.manager.load("CodeFormers")
        assert SharedModelManager.manager.codeformer.is_model_loaded("CodeFormers") is True
        data = {
            "model": "CodeFormers",
            "source_image": Image.open("images/test_facefix.png"),
        }
        pil_image = self.horde.image_facefix(data)
        assert pil_image is not None
        width, height = pil_image.size
        pil_image.save("images/image_facefix_codeformers.webp", quality=90)

    @pytest.mark.mm_model("gfpgan")
    def test_image_facefix_gfpgan(self):
        SharedModelManager.manager.load("GFPGAN")
        assert SharedModelManager.manager.gfpgan.is_model_loaded("GFPGAN") is True
        data = {
            "model": "GFPGAN",
            "source_image": Image.open("images/test_facefix.png"),
        }
        pil_image = self.horde.image_facefix(data)
        assert pil_image is not None
        width, height = pil_image.size
        pil_image.save("images/image_facefix_gfpgan.webp", quality=90)



# test_horde.py
import pytest
from PIL import Image

from hordelib.horde import HordeLib
from hordelib.shared_model_manager import SharedModelManager


class TestHordeInference:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.horde = HordeLib()

        self.default_model_manager_args = {
            # aitemplate
            # "blip": True,
            # "clip": True,
            # "codeformer": True,
            "compvis": True,
            "controlnet": True,
            # "diffusers": True,
            # "esrgan": True,
            # "gfpgan": True,
            # "safety_checker": True,
        }
        SharedModelManager.loadModelManagers(**self.default_model_manager_args)
        assert SharedModelManager.manager is not None
        SharedModelManager.manager.load("Deliberate")
        for preproc in HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys():
            SharedModelManager.manager.controlnet.download_control_type(preproc)
        yield
        del self.horde
        SharedModelManager._instance = None
        SharedModelManager.manager = None

    def test_controlnet_sd1(self):
        data = {
            "sampler_name": "k_dpmpp_2m",
            "cfg_scale": 7.5,
            "denoising_strength": 1.0,
            "seed": 123456789,
            "height": 512,
            "width": 512,
            "karras": True,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": "",
            "image_is_control": False,
            "return_control_map": False,
            "prompt": "a man walking in the snow",
            "ddim_steps": 25,
            "n_iter": 1,
            "model": "Deliberate",
            "source_image": Image.open("images/test_db0.jpg"),
            "source_processing": "img2img",
        }
        assert self.horde is not None

        for preproc in HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys():
            if preproc == "scribble" or preproc == "mlsd":
                # Skip
                continue
            assert (
                SharedModelManager.manager.controlnet.check_control_type_available(
                    preproc,
                    "stable diffusion 1",
                )
                is True
            )
            data["control_type"] = preproc
            pil_image = self.horde.basic_inference(data)
            assert pil_image is not None
            pil_image.save(f"images/controlnet_{preproc}.webp", quality=90)

    def test_controlnet_strength(self):
        data = {
            "sampler_name": "k_dpmpp_2m",
            "cfg_scale": 7.5,
            "denoising_strength": 1.0,
            "seed": 123456789,
            "height": 512,
            "width": 512,
            "karras": True,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": "canny",
            "image_is_control": False,
            "return_control_map": False,
            "prompt": "a man walking on the moon",
            "ddim_steps": 25,
            "n_iter": 1,
            "model": "Deliberate",
            "source_image": Image.open("images/test_db0.jpg"),
            "source_processing": "img2img",
        }
        for strength in [1.0, 0.5, 0.2]:
            data["control_strength"] = strength
            pil_image = self.horde.basic_inference(data)
            assert pil_image is not None
            pil_image.save(f"images/controlnet_strength_{strength}.webp", quality=90)

    def test_controlnet_hires_fix(self):
        data = {
            "sampler_name": "k_dpmpp_2m",
            "cfg_scale": 7.5,
            "denoising_strength": 1.0,
            "seed": 1234345378856789,
            "height": 768,
            "width": 768,
            "karras": True,
            "tiling": False,
            "hires_fix": True,
            "hires_fix_denoising_strength": 0.0,
            "clip_skip": 1,
            "control_type": "canny",
            "image_is_control": False,
            "return_control_map": False,
            "prompt": "a man walking in the jungle",
            "ddim_steps": 15,
            "n_iter": 1,
            "model": "Deliberate",
            "source_image": Image.open("images/test_db0.jpg"),
            "source_processing": "img2img",
        }
        for denoise in [0.4, 0.5, 0.6]:
            data["hires_fix_denoising_strength"] = denoise
            pil_image = self.horde.basic_inference(data)
            assert pil_image is not None
            pil_image.save(f"images/controlnet_hires_fix_denoise_{denoise}.webp", quality=90)



import typing

import torch
from diffusers.pipelines.stable_diffusion.safety_checker import (
    StableDiffusionSafetyChecker,
)
from loguru import logger
from typing_extensions import override

from hordelib.consts import MODEL_CATEGORY_NAMES, MODEL_DB_NAMES
from hordelib.model_manager.base import BaseModelManager


class SafetyCheckerModelManager(BaseModelManager):
    def __init__(self, download_reference=False):
        super().__init__(
            models_db_name=MODEL_DB_NAMES[MODEL_CATEGORY_NAMES.safety_checker],
            download_reference=download_reference,
        )

    @override
    def modelToRam(
        self,
        model_name: str,
        half_precision=True,
        gpu_id=0,
        cpu_only=True,
        **kwargs,
    ) -> dict[str, typing.Any]:
        if not self.cuda_available:
            cpu_only = True
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        logger.info(f"Loading model {model_name} on {device}")
        logger.info(f"Model path: {self.modelFolderPath}")
        model = StableDiffusionSafetyChecker.from_pretrained(self.modelFolderPath)
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        return {"model": model, "device": device, "half_precision": half_precision}



import pytest

from hordelib.horde import HordeLib
from hordelib.model_manager.lora import LoraModelManager
from hordelib.utils.logger import HordeLog


class TestModelManagerLora:
    horde = HordeLib()
    default_model_manager_args: dict

    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.horde = HordeLib()
        HordeLog.initialise(True)
        HordeLog.set_logger_verbosity(5)
        HordeLog.quiesce_logger(0)

        yield
        del self.horde

    def test_downloading_default_sync(self):
        download_amount = 1024
        mml = LoraModelManager(
            allowed_top_lora_storage=download_amount,
            download_wait=True,
        )
        mml.download_default_loras()
        assert mml.are_downloads_complete() is True
        assert mml.calculate_downloaded_loras() > download_amount
        assert mml.calculate_downloaded_loras() < download_amount * 1.4

    def test_downloading_default_async(self):
        download_amount = 1024
        mml = LoraModelManager(
            allowed_top_lora_storage=download_amount,
            download_wait=False,
        )
        mml.download_default_loras()
        assert mml.are_downloads_complete() is False
        mml.wait_for_downloads()
        assert mml.are_downloads_complete() is True
        assert mml.calculate_downloaded_loras() > download_amount
        assert mml.calculate_downloaded_loras() < download_amount * 1.4

    def test_fuzzy_search(self):
        download_amount = 1024
        mml = LoraModelManager(
            allowed_top_lora_storage=download_amount,
            download_wait=False,
        )
        mml.download_default_loras()
        mml.wait_for_downloads()
        assert mml.fuzzy_find_lora("Glowing Runes") == "glowingrunesai"
        assert mml.fuzzy_find_lora("Glowing Robots") is None
        assert mml.fuzzy_find_lora("GlowingRobots") is None
        assert mml.fuzzy_find_lora("GlowingRobotsAI") is None

    def test_lora_search(self):
        download_amount = 1024
        mml = LoraModelManager(
            allowed_top_lora_storage=download_amount,
            download_wait=False,
        )
        mml.download_default_loras()
        mml.wait_for_downloads()
        assert mml.get_lora_name("GlowingRunesAI") == "GlowingRunesAI"
        assert mml.get_lora_name("GlowingRunes") == "GlowingRunesAI"
        assert mml.get_lora_name("Glowing Runes") == "GlowingRunesAI"
        assert len(mml.get_lora_triggers("GlowingRunesAI")) > 1
        # We can't rely on triggers not changing
        assert mml.find_lora_trigger("GlowingRunesAI", "blue") is not None
        assert "blue" in mml.find_lora_trigger("GlowingRunesAI", "blue").lower()
        assert "red" in mml.find_lora_trigger("GlowingRunesAI", "red").lower()
        assert mml.find_lora_trigger("GlowingRunesAI", "pale blue") is None  # This is too much to fuzz
        assert mml.get_lora_name("Dra9onScaleAI") is not None
        assert mml.get_lora_name("DragonScale") is not None
        assert mml.get_lora_name("Dragon Scale AI") is not None
        assert mml.find_lora_trigger("Dra9onScaleAI", "Dr490nSc4leAI") is not None
        assert mml.find_lora_trigger("DragonScale", "DragonScaleAI") is not None



import pytest

import hordelib


class TestSafetyChecker:
    _initialised = False

    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        if not self._initialised:
            hordelib.initialise()
        from hordelib.model_manager.safety_checker import SafetyCheckerModelManager

        self.safety_model_manager = SafetyCheckerModelManager()
        assert self.safety_model_manager is not None
        yield
        del self.safety_model_manager

    def test_safety_checker_load_defaults(self):
        success = self.safety_model_manager.load("safety_checker", cpu_only=True)
        assert success

    # def test_safety_checker_load_cpu_only(self):
    #     success = self.safety_model_manager.load("safety_checker", cpu_only=True)
    #     assert success



# test_horde.py
import pytest

from hordelib.consts import EXCLUDED_MODEL_NAMES
from hordelib.horde import HordeLib
from hordelib.shared_model_manager import SharedModelManager
from hordelib.utils.logger import HordeLog


class TestSharedModelManager:
    horde = HordeLib()
    default_model_manager_args: dict

    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.horde = HordeLib()
        HordeLog.initialise(True)
        HordeLog.set_logger_verbosity(5)
        HordeLog.quiesce_logger(0)

        self.default_model_manager_args = {  # XXX # TODO
            # aitemplate
            "blip": True,
            "clip": True,
            "codeformer": True,
            "compvis": True,
            "controlnet": True,
            # "diffusers": True,
            "esrgan": True,
            "gfpgan": True,
            "safety_checker": True,
        }
        SharedModelManager.loadModelManagers(**self.default_model_manager_args)
        assert SharedModelManager.manager is not None
        yield
        del self.horde
        SharedModelManager._instance = None
        SharedModelManager.manager = None

    def test_singleton(self):
        a = SharedModelManager()
        b = SharedModelManager()
        assert a.manager is b.manager

    def test_horde_model_manager_init(self):
        assert SharedModelManager.manager is not None
        # assert SharedModelManager.manager.aitemplate is not None # XXX # FIXME
        assert SharedModelManager.manager.blip is not None  # XXX
        assert SharedModelManager.manager.clip is not None
        assert SharedModelManager.manager.codeformer is not None
        assert SharedModelManager.manager.compvis is not None
        assert SharedModelManager.manager.controlnet is not None
        # assert SharedModelManager.manager.diffusers is not None
        assert SharedModelManager.manager.esrgan is not None
        assert SharedModelManager.manager.gfpgan is not None
        assert SharedModelManager.manager.safety_checker is not None

    def test_horde_model_manager_reload_db(self):
        assert SharedModelManager.manager is not None
        SharedModelManager.manager.reload_database()

    def test_horde_model_manager_download_model(self):
        assert SharedModelManager.manager is not None
        result: bool | None = SharedModelManager.manager.download_model("Deliberate")
        assert result is True

    def test_horde_model_manager_validate(self):
        assert SharedModelManager.manager is not None
        SharedModelManager.manager.load("Deliberate")
        result: bool | None = SharedModelManager.manager.validate_model("Deliberate")
        assert result is True

    def test_taint_models(self):
        assert SharedModelManager.manager is not None
        SharedModelManager.manager.taint_models(["Deliberate"])
        assert "Deliberate" not in SharedModelManager.manager.get_available_models()
        assert SharedModelManager.manager.is_model_loaded("Deliberate") is False

    # XXX add a test for model missing?
    def test_horde_model_manager_unload_model(self):
        assert SharedModelManager.manager is not None
        SharedModelManager.manager.load("Deliberate")
        assert SharedModelManager.manager.is_model_loaded("Deliberate") is True
        result = SharedModelManager.manager.unload_model("Deliberate")
        assert result is True
        assert SharedModelManager.manager.is_model_loaded("Deliberate") is False

    def test_model_load_checking(self):
        assert SharedModelManager.manager is not None
        assert SharedModelManager.manager.is_model_loaded("Deliberate") is False
        assert SharedModelManager.manager.is_model_loaded("GFPGAN") is False
        assert SharedModelManager.manager.is_model_loaded("RealESRGAN_x4plus") is False
        assert SharedModelManager.manager.is_model_loaded("4x_NMKD_Superscale_SP") is False
        SharedModelManager.manager.load("Deliberate")
        SharedModelManager.manager.load("GFPGAN")
        SharedModelManager.manager.load("RealESRGAN_x4plus")
        SharedModelManager.manager.load("4x_NMKD_Superscale_SP")
        assert SharedModelManager.manager.is_model_loaded("Deliberate") is True
        assert SharedModelManager.manager.is_model_loaded("GFPGAN") is True
        assert SharedModelManager.manager.is_model_loaded("RealESRGAN_x4plus") is True
        assert SharedModelManager.manager.is_model_loaded("4x_NMKD_Superscale_SP") is True

    def test_model_excluding(self):
        assert SharedModelManager.manager is not None
        for excluded_model in EXCLUDED_MODEL_NAMES:
            assert not SharedModelManager.manager.load(excluded_model)

    def test_check_sha(self):
        """Check the sha256 hashes of all models. If the .sha file doesn't exist, this will write it out."""
        assert SharedModelManager.manager is not None
        for model_manager in SharedModelManager.manager.active_model_managers:
            for model in model_manager.available_models:
                model_file_details = model_manager.get_model_files(model)
                for file in model_file_details:
                    path = file.get("path")
                    if not (".pt" in path or ".ckpt" in path or ".safetensors" in path):
                        continue
                    model_manager.get_file_sha256_hash(f"{model_manager.modelFolderPath}/{path}")
        pass

    def test_check_validate_all_available_models(self):
        assert SharedModelManager.manager is not None
        for model_manager in SharedModelManager.manager.active_model_managers:
            for model in model_manager.available_models:
                assert model_manager.validate_model(model)

    def test_preload_annotators(self):
        assert SharedModelManager.preloadAnnotators()



import typing

from typing_extensions import override

from hordelib.comfy_horde import horde_load_checkpoint
from hordelib.consts import MODEL_CATEGORY_NAMES, MODEL_DB_NAMES
from hordelib.model_manager.base import BaseModelManager


class DiffusersModelManager(BaseModelManager):
    def __init__(self, download_reference=False):
        raise NotImplementedError("Diffusers are not yet supported")

        super().__init__(
            models_db_name=MODEL_DB_NAMES[MODEL_CATEGORY_NAMES.diffusers],
            download_reference=download_reference,
        )

    @override
    def modelToRam(
        self,
        model_name: str,
        **kwargs,
    ) -> dict[str, typing.Any]:
        raise NotImplementedError("Diffusers are not yet supported")



import time
import typing
from pathlib import Path

import clip
import open_clip
import torch
from loguru import logger
from typing_extensions import override

from hordelib.config_path import get_hordelib_path
from hordelib.consts import MODEL_CATEGORY_NAMES, MODEL_DB_NAMES
from hordelib.model_manager.base import BaseModelManager


class ClipModelManager(BaseModelManager):
    def __init__(self, download_reference=False):
        super().__init__(
            models_db_name=MODEL_DB_NAMES[MODEL_CATEGORY_NAMES.clip],
            download_reference=download_reference,
        )

    def load_ranking_lists(self):
        ranking_lists = {}
        ranking_lists_path = Path(get_hordelib_path()).joinpath(
            "clip/",
            "ranking_lists/",
        )
        for file in ranking_lists_path.glob("*.txt"):
            ranking_lists[file.stem] = load_list(file)
        return ranking_lists

    def load_coca(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        model_path = self.get_model_files(model_name)[0]["path"]
        model_path = f"{self.modelFolderPath}/{model_path}"
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, _, transform = open_clip.create_model_and_transforms(
            "coca_ViT-L-14",
            pretrained=model_path,
            device=device,
            precision="fp16" if half_precision else "fp32",
        )
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        return {
            "model": model,
            "device": device,
            "transform": transform,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    def load_open_clip(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        pretrained = self.get_model(model_name)["pretrained_name"]
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, _, preprocess = open_clip.create_model_and_transforms(
            model_name,
            pretrained=pretrained,
            cache_dir=self.modelFolderPath,
            device=device,
            precision="fp16" if half_precision else "fp32",
        )
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        ranking_lists = self.load_ranking_lists()
        return {
            "model": model,
            "device": device,
            "preprocess": preprocess,
            "ranking_lists": ranking_lists,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    def load_clip(self, model_name, half_precision=True, gpu_id=0, cpu_only=False):
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        model, preprocess = clip.load(
            model_name,
            device=device,
            download_root=self.modelFolderPath,
        )
        model = model.eval()
        if half_precision:
            model = model.half()
        ranking_lists = self.load_ranking_lists()
        return {
            "model": model,
            "device": device,
            "preprocess": preprocess,
            "ranking_lists": ranking_lists,
            "half_precision": half_precision,
            "cache_name": model_name.replace("/", "_"),
        }

    @override
    def modelToRam(
        self,
        model_name: str,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        **kwargs,
    ) -> dict[str, typing.Any]:
        """
        model_name: str. Name of the model to load. See available_models for a list of available models.
        half_precision: bool. If True, the model will be loaded in half precision.
        gpu_id: int. The id of the gpu to use. If the gpu is not available, the model will be loaded on the cpu.
        cpu_only: bool. If True, the model will be loaded on the cpu. If True, half_precision will be set to False.
        """
        loaded_model_info = None
        if not self.cuda_available:
            cpu_only = True
        tic = time.time()
        logger.init(f"{model_name}", status="Loading")  # logger.init
        if self.model_reference[model_name]["type"] == "open_clip":
            loaded_model_info = self.load_open_clip(
                model_name,
                half_precision,
                gpu_id,
                cpu_only,
            )
            self.add_loaded_model(model_name, loaded_model_info)
        elif self.model_reference[model_name]["type"] == "clip":
            loaded_model_info = self.load_clip(
                model_name,
                half_precision,
                gpu_id,
                cpu_only,
            )
            self.add_loaded_model(model_name, loaded_model_info)
        elif self.model_reference[model_name]["type"] == "coca":
            loaded_model_info = self.load_coca(
                model_name,
                half_precision,
                gpu_id,
                cpu_only,
            )
            self.add_loaded_models(model_name, loaded_model_info)
        else:
            logger.error(
                f"Unknown model type: {self.model_reference[model_name]['type']}",
            )
            return {}  # XXX # FIXME
        if not loaded_model_info:
            logger.init_err(f"Failed to load {model_name}", status="Error")
            return {}  # XXX # FIXME

        logger.init_ok(f"Loading {model_name}", status="Success")  # logger.init_ok
        toc = time.time()
        logger.init_ok(
            f"Loading {model_name}: Took {toc-tic} seconds",
            status="Success",
        )  # logger.init_ok
        return loaded_model_info


def load_list(filename):
    with open(filename, encoding="utf-8", errors="replace") as f:
        return [line.strip() for line in f.readlines()]



import os
import sys
import asyncio
import nodes
import folder_paths
import execution
import uuid
import json
import glob
from PIL import Image
from io import BytesIO

try:
    import aiohttp
    from aiohttp import web
except ImportError:
    print("Module 'aiohttp' not installed. Please install it via:")
    print("pip install aiohttp")
    print("or")
    print("pip install -r requirements.txt")
    sys.exit()

import mimetypes
from comfy.cli_args import args


@web.middleware
async def cache_control(request: web.Request, handler):
    response: web.Response = await handler(request)
    if request.path.endswith('.js') or request.path.endswith('.css'):
        response.headers.setdefault('Cache-Control', 'no-cache')
    return response

def create_cors_middleware(allowed_origin: str):
    @web.middleware
    async def cors_middleware(request: web.Request, handler):
        if request.method == "OPTIONS":
            # Pre-flight request. Reply successfully:
            response = web.Response()
        else:
            response = await handler(request)

        response.headers['Access-Control-Allow-Origin'] = allowed_origin
        response.headers['Access-Control-Allow-Methods'] = 'POST, GET, DELETE, PUT, OPTIONS'
        response.headers['Access-Control-Allow-Headers'] = 'Content-Type, Authorization'
        response.headers['Access-Control-Allow-Credentials'] = 'true'
        return response

    return cors_middleware

class PromptServer():
    def __init__(self, loop):
        PromptServer.instance = self

        mimetypes.init(); 
        mimetypes.types_map['.js'] = 'application/javascript; charset=utf-8'
        self.prompt_queue = None
        self.loop = loop
        self.messages = asyncio.Queue()
        self.number = 0

        middlewares = [cache_control]
        if args.enable_cors_header:
            middlewares.append(create_cors_middleware(args.enable_cors_header))

        self.app = web.Application(client_max_size=20971520, middlewares=middlewares)
        self.sockets = dict()
        self.web_root = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), "web")
        routes = web.RouteTableDef()
        self.routes = routes
        self.last_node_id = None
        self.client_id = None

        @routes.get('/ws')
        async def websocket_handler(request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            sid = request.rel_url.query.get('clientId', '')
            if sid:
                # Reusing existing session, remove old
                self.sockets.pop(sid, None)
            else:
                sid = uuid.uuid4().hex

            self.sockets[sid] = ws

            try:
                # Send initial state to the new client
                await self.send("status", { "status": self.get_queue_info(), 'sid': sid }, sid)
                # On reconnect if we are the currently executing client send the current node
                if self.client_id == sid and self.last_node_id is not None:
                    await self.send("executing", { "node": self.last_node_id }, sid)
                    
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.ERROR:
                        print('ws connection closed with exception %s' % ws.exception())
            finally:
                self.sockets.pop(sid, None)
            return ws

        @routes.get("/")
        async def get_root(request):
            return web.FileResponse(os.path.join(self.web_root, "index.html"))

        @routes.get("/embeddings")
        def get_embeddings(self):
            embeddings = folder_paths.get_filename_list("embeddings")
            return web.json_response(list(map(lambda a: os.path.splitext(a)[0].lower(), embeddings)))

        @routes.get("/extensions")
        async def get_extensions(request):
            files = glob.glob(os.path.join(self.web_root, 'extensions/**/*.js'), recursive=True)
            return web.json_response(list(map(lambda f: "/" + os.path.relpath(f, self.web_root).replace("\\", "/"), files)))

        def get_dir_by_type(dir_type):
            if dir_type is None:
                dir_type = "input"

            if dir_type == "input":
                type_dir = folder_paths.get_input_directory()
            elif dir_type == "temp":
                type_dir = folder_paths.get_temp_directory()
            elif dir_type == "output":
                type_dir = folder_paths.get_output_directory()

            return type_dir, dir_type

        def image_upload(post, image_save_function=None):
            image = post.get("image")
            overwrite = post.get("overwrite")

            image_upload_type = post.get("type")
            upload_dir, image_upload_type = get_dir_by_type(image_upload_type)

            if image and image.file:
                filename = image.filename
                if not filename:
                    return web.Response(status=400)

                subfolder = post.get("subfolder", "")
                full_output_folder = os.path.join(upload_dir, os.path.normpath(subfolder))

                if os.path.commonpath((upload_dir, os.path.abspath(full_output_folder))) != upload_dir:
                    return web.Response(status=400)

                if not os.path.exists(full_output_folder):
                    os.makedirs(full_output_folder)

                split = os.path.splitext(filename)
                filepath = os.path.join(full_output_folder, filename)

                if overwrite is not None and (overwrite == "true" or overwrite == "1"):
                    pass
                else:
                    i = 1
                    while os.path.exists(filepath):
                        filename = f"{split[0]} ({i}){split[1]}"
                        filepath = os.path.join(full_output_folder, filename)
                        i += 1

                if image_save_function is not None:
                    image_save_function(image, post, filepath)
                else:
                    with open(filepath, "wb") as f:
                        f.write(image.file.read())

                return web.json_response({"name" : filename, "subfolder": subfolder, "type": image_upload_type})
            else:
                return web.Response(status=400)

        @routes.post("/upload/image")
        async def upload_image(request):
            post = await request.post()
            return image_upload(post)

        @routes.post("/upload/mask")
        async def upload_mask(request):
            post = await request.post()

            def image_save_function(image, post, filepath):
                original_pil = Image.open(post.get("original_image").file).convert('RGBA')
                mask_pil = Image.open(image.file).convert('RGBA')

                # alpha copy
                new_alpha = mask_pil.getchannel('A')
                original_pil.putalpha(new_alpha)
                original_pil.save(filepath, compress_level=4)

            return image_upload(post, image_save_function)

        @routes.get("/view")
        async def view_image(request):
            if "filename" in request.rel_url.query:
                filename = request.rel_url.query["filename"]
                filename,output_dir = folder_paths.annotated_filepath(filename)

                # validation for security: prevent accessing arbitrary path
                if filename[0] == '/' or '..' in filename:
                    return web.Response(status=400)

                if output_dir is None:
                    type = request.rel_url.query.get("type", "output")
                    output_dir = folder_paths.get_directory_by_type(type)

                if output_dir is None:
                    return web.Response(status=400)

                if "subfolder" in request.rel_url.query:
                    full_output_dir = os.path.join(output_dir, request.rel_url.query["subfolder"])
                    if os.path.commonpath((os.path.abspath(full_output_dir), output_dir)) != output_dir:
                        return web.Response(status=403)
                    output_dir = full_output_dir

                filename = os.path.basename(filename)
                file = os.path.join(output_dir, filename)

                if os.path.isfile(file):
                    if 'channel' not in request.rel_url.query:
                        channel = 'rgba'
                    else:
                        channel = request.rel_url.query["channel"]

                    if channel == 'rgb':
                        with Image.open(file) as img:
                            if img.mode == "RGBA":
                                r, g, b, a = img.split()
                                new_img = Image.merge('RGB', (r, g, b))
                            else:
                                new_img = img.convert("RGB")

                            buffer = BytesIO()
                            new_img.save(buffer, format='PNG')
                            buffer.seek(0)

                            return web.Response(body=buffer.read(), content_type='image/png',
                                                headers={"Content-Disposition": f"filename=\"{filename}\""})

                    elif channel == 'a':
                        with Image.open(file) as img:
                            if img.mode == "RGBA":
                                _, _, _, a = img.split()
                            else:
                                a = Image.new('L', img.size, 255)

                            # alpha img
                            alpha_img = Image.new('RGBA', img.size)
                            alpha_img.putalpha(a)
                            alpha_buffer = BytesIO()
                            alpha_img.save(alpha_buffer, format='PNG')
                            alpha_buffer.seek(0)

                            return web.Response(body=alpha_buffer.read(), content_type='image/png',
                                                headers={"Content-Disposition": f"filename=\"{filename}\""})
                    else:
                        return web.FileResponse(file, headers={"Content-Disposition": f"filename=\"{filename}\""})

            return web.Response(status=404)

        @routes.get("/prompt")
        async def get_prompt(request):
            return web.json_response(self.get_queue_info())

        @routes.get("/object_info")
        async def get_object_info(request):
            out = {}
            for x in nodes.NODE_CLASS_MAPPINGS:
                obj_class = nodes.NODE_CLASS_MAPPINGS[x]
                info = {}
                info['input'] = obj_class.INPUT_TYPES()
                info['output'] = obj_class.RETURN_TYPES
                info['output_is_list'] = obj_class.OUTPUT_IS_LIST if hasattr(obj_class, 'OUTPUT_IS_LIST') else [False] * len(obj_class.RETURN_TYPES)
                info['output_name'] = obj_class.RETURN_NAMES if hasattr(obj_class, 'RETURN_NAMES') else info['output']
                info['name'] = x
                info['display_name'] = nodes.NODE_DISPLAY_NAME_MAPPINGS[x] if x in nodes.NODE_DISPLAY_NAME_MAPPINGS.keys() else x
                info['description'] = ''
                info['category'] = 'sd'
                if hasattr(obj_class, 'CATEGORY'):
                    info['category'] = obj_class.CATEGORY
                out[x] = info
            return web.json_response(out)

        @routes.get("/history")
        async def get_history(request):
            return web.json_response(self.prompt_queue.get_history())

        @routes.get("/queue")
        async def get_queue(request):
            queue_info = {}
            current_queue = self.prompt_queue.get_current_queue()
            queue_info['queue_running'] = current_queue[0]
            queue_info['queue_pending'] = current_queue[1]
            return web.json_response(queue_info)

        @routes.post("/prompt")
        async def post_prompt(request):
            print("got prompt")
            resp_code = 200
            out_string = ""
            json_data =  await request.json()

            if "number" in json_data:
                number = float(json_data['number'])
            else:
                number = self.number
                if "front" in json_data:
                    if json_data['front']:
                        number = -number

                self.number += 1

            if "prompt" in json_data:
                prompt = json_data["prompt"]
                valid = execution.validate_prompt(prompt)
                extra_data = {}
                if "extra_data" in json_data:
                    extra_data = json_data["extra_data"]

                if "client_id" in json_data:
                    extra_data["client_id"] = json_data["client_id"]
                if valid[0]:
                    prompt_id = str(uuid.uuid4())
                    self.prompt_queue.put((number, prompt_id, prompt, extra_data, valid[2]))
                    return web.json_response({"prompt_id": prompt_id})
                else:
                    print("invalid prompt:", valid[1])
                    return web.json_response({"error": valid[1]}, status=400)
            else:
                return web.json_response({"error": "no prompt"}, status=400)

        @routes.post("/queue")
        async def post_queue(request):
            json_data =  await request.json()
            if "clear" in json_data:
                if json_data["clear"]:
                    self.prompt_queue.wipe_queue()
            if "delete" in json_data:
                to_delete = json_data['delete']
                for id_to_delete in to_delete:
                    delete_func = lambda a: a[1] == id_to_delete
                    self.prompt_queue.delete_queue_item(delete_func)

            return web.Response(status=200)

        @routes.post("/interrupt")
        async def post_interrupt(request):
            nodes.interrupt_processing()
            return web.Response(status=200)

        @routes.post("/history")
        async def post_history(request):
            json_data =  await request.json()
            if "clear" in json_data:
                if json_data["clear"]:
                    self.prompt_queue.wipe_history()
            if "delete" in json_data:
                to_delete = json_data['delete']
                for id_to_delete in to_delete:
                    self.prompt_queue.delete_history_item(id_to_delete)

            return web.Response(status=200)
        
    def add_routes(self):
        self.app.add_routes(self.routes)
        self.app.add_routes([
            web.static('/', self.web_root, follow_symlinks=True),
        ])

    def get_queue_info(self):
        prompt_info = {}
        exec_info = {}
        exec_info['queue_remaining'] = self.prompt_queue.get_tasks_remaining()
        prompt_info['exec_info'] = exec_info
        return prompt_info

    async def send(self, event, data, sid=None):
        message = {"type": event, "data": data}
       
        if isinstance(message, str) == False:
            message = json.dumps(message)

        if sid is None:
            for ws in self.sockets.values():
                await ws.send_str(message)
        elif sid in self.sockets:
            await self.sockets[sid].send_str(message)

    def send_sync(self, event, data, sid=None):
        self.loop.call_soon_threadsafe(
            self.messages.put_nowait, (event, data, sid))

    def queue_updated(self):
        self.send_sync("status", { "status": self.get_queue_info() })

    async def publish_loop(self):
        while True:
            msg = await self.messages.get()
            await self.send(*msg)

    async def start(self, address, port, verbose=True, call_on_start=None):
        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, address, port)
        await site.start()

        if address == '':
            address = '0.0.0.0'
        if verbose:
            print("Starting server\n")
            print("To see the GUI go to: http://{}:{}".format(address, port))
        if call_on_start is not None:
            call_on_start(address, port)




import importlib.resources as importlib_resources
import time
import typing
from pathlib import Path

import torch
from loguru import logger
from typing_extensions import override

from hordelib.config_path import get_hordelib_path
from hordelib.consts import MODEL_CATEGORY_NAMES, MODEL_DB_NAMES
from hordelib.model_manager.base import BaseModelManager
from hordelib.utils.blip.blip import blip_decoder


class BlipModelManager(BaseModelManager):
    def __init__(self, download_reference=False):
        super().__init__(
            models_db_name=MODEL_DB_NAMES[MODEL_CATEGORY_NAMES.blip],
            download_reference=download_reference,
        )

    @override
    def modelToRam(
        self,
        model_name,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        blip_image_eval_size=512,
        **kwargs,
    ) -> dict[str, typing.Any]:
        if not self.cuda_available:
            cpu_only = True
        vit = "base" if model_name == "BLIP" else "large"
        model_path = self.get_model_files(model_name)[0]["path"]
        model_path = f"{self.modelFolderPath}/{model_path}"
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        logger.info(f"Loading model {model_name} on {device}")
        logger.info(f"Model path: {model_path}")
        med_path = Path(get_hordelib_path()).joinpath(
            "model_database/",
            "med_config.json",
        )
        with importlib_resources.as_file(med_path.resolve()) as med_config:
            logger.info(f"Med config path: {med_config}")
            model = blip_decoder(
                pretrained=model_path,
                med_config=med_config,
                image_size=blip_image_eval_size,
                vit=vit,
            )
        model = model.eval()
        model.to(device)
        if half_precision:
            model = model.half()
        return {"model": model, "device": device, "half_precision": half_precision}



# node_model_loader.py
# Simple proof of concept custom node to load models.

import contextlib
import os
import pickle
import time

from loguru import logger


class HordeCheckpointLoader:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model_manager": ("<model manager instance>",),
                "model_name": ("<model name>",),
            },
        }

    RETURN_TYPES = ("MODEL", "CLIP", "VAE")
    FUNCTION = "load_checkpoint"

    CATEGORY = "loaders"

    def load_checkpoint(
        self,
        model_manager,
        model_name,
    ):
        logger.debug(f"Loading model {model_name} through our custom node")

        if model_manager.manager is None:
            logger.error("horde_model_manager appears to be missing!")
            raise RuntimeError  # XXX better guarantees need to be made

        loaded_models = model_manager.manager.loaded_models
        if model_name not in loaded_models:
            logger.error(f"Model {model_name} is not loaded")
            raise RuntimeError  # XXX better guarantees need to be made

        model = loaded_models[model_name]["model"]
        clip = loaded_models[model_name]["clip"]
        vae = loaded_models[model_name]["vae"]

        # If we got strings, not objects, it's a cache reference, load the cache
        if type(model) is str:
            start_time = time.time()
            logger.info(f"Loading from disk cache model {model_name}")
            model_cache = model
            try:
                with model_manager.manager.disk_read_mutex:
                    with open(model, "rb") as cache:
                        model = pickle.load(cache)
                        vae = pickle.load(cache)
                        clip = pickle.load(cache)
                # Record this model as being in ram again
                model_manager.manager.move_from_disk_cache(model_name, model, clip, vae)
                logger.info(
                    f"Loaded model {model_name} from disk cache in {round(time.time() - start_time, 1)} seconds",
                )
            except (pickle.PickleError, EOFError):
                # Most likely corrupt cache file, remove the file
                with contextlib.suppress(OSError):
                    os.remove(model)  # ... at least try to remove it

                raise Exception(f"Model cache file {model_cache} was corrupt. It has been removed.")

        # XXX # TODO I would like to revisit this dict->tuple conversion at some point soon
        return (model, clip, vae)


NODE_CLASS_MAPPINGS = {"HordeCheckpointLoader": HordeCheckpointLoader}



import asyncio
import itertools
import os
import shutil
import threading

from comfy.cli_args import args
import comfy.utils

if os.name == "nt":
    import logging
    logging.getLogger("xformers").addFilter(lambda record: 'A matching Triton is not available' not in record.getMessage())

if __name__ == "__main__":
    if args.dont_upcast_attention:
        print("disabling upcasting of attention")
        os.environ['ATTN_PRECISION'] = "fp16"

    if args.cuda_device is not None:
        os.environ['CUDA_VISIBLE_DEVICES'] = str(args.cuda_device)
        print("Set cuda device to:", args.cuda_device)


import yaml

import execution
import folder_paths
import server
from nodes import init_custom_nodes, load_custom_nodes


def prompt_worker(q, server):
    e = execution.PromptExecutor(server)
    while True:
        item, item_id = q.get()
        e.execute(item[2], item[1], item[3], item[4])
        q.task_done(item_id, e.outputs)

async def run(server, address='', port=8188, verbose=True, call_on_start=None):
    await asyncio.gather(server.start(address, port, verbose, call_on_start), server.publish_loop())

def hijack_progress(server):
    def hook(value, total):
        server.send_sync("progress", { "value": value, "max": total}, server.client_id)
    comfy.utils.set_progress_bar_global_hook(hook)

def cleanup_temp():
    temp_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "temp")
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir, ignore_errors=True)

def load_extra_path_config(yaml_path):
    with open(yaml_path, 'r') as stream:
        config = yaml.safe_load(stream)
    for c in config:
        conf = config[c]
        if conf is None:
            continue
        base_path = None
        if "base_path" in conf:
            base_path = conf.pop("base_path")
        for x in conf:
            for y in conf[x].split("\n"):
                if len(y) == 0:
                    continue
                full_path = y
                if base_path is not None:
                    full_path = os.path.join(base_path, full_path)
                print("Adding extra search path", x, full_path)
                folder_paths.add_model_folder_path(x, full_path)

if __name__ == "__main__":
    cleanup_temp()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    server = server.PromptServer(loop)
    q = execution.PromptQueue(server)

    extra_model_paths_config_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "extra_model_paths.yaml")
    if os.path.isfile(extra_model_paths_config_path):
        load_extra_path_config(extra_model_paths_config_path)

    if args.extra_model_paths_config:
        for config_path in itertools.chain(*args.extra_model_paths_config):
            load_extra_path_config(config_path)

    init_custom_nodes()
    server.add_routes()
    hijack_progress(server)

    threading.Thread(target=prompt_worker, daemon=True, args=(q,server,)).start()

    if args.output_directory:
        output_dir = os.path.abspath(args.output_directory)
        print(f"Setting output directory to: {output_dir}")
        folder_paths.set_output_directory(output_dir)

    if args.quick_test_for_ci:
        exit(0)

    call_on_start = None
    if args.auto_launch:
        def startup_server(address, port):
            import webbrowser
            webbrowser.open("http://{}:{}".format(address, port))
        call_on_start = startup_server

    if os.name == "nt":
        try:
            loop.run_until_complete(run(server, address=args.listen, port=args.port, verbose=not args.dont_print_server, call_on_start=call_on_start))
        except KeyboardInterrupt:
            pass
    else:
        loop.run_until_complete(run(server, address=args.listen, port=args.port, verbose=not args.dont_print_server, call_on_start=call_on_start))

    cleanup_temp()



import comfy.utils
import folder_paths
import torch

def load_hypernetwork_patch(path, strength):
    sd = comfy.utils.load_torch_file(path, safe_load=True)
    activation_func = sd.get('activation_func', 'linear')
    is_layer_norm = sd.get('is_layer_norm', False)
    use_dropout = sd.get('use_dropout', False)
    activate_output = sd.get('activate_output', False)
    last_layer_dropout = sd.get('last_layer_dropout', False)

    valid_activation = {
        "linear": torch.nn.Identity,
        "relu": torch.nn.ReLU,
        "leakyrelu": torch.nn.LeakyReLU,
        "elu": torch.nn.ELU,
        "swish": torch.nn.Hardswish,
        "tanh": torch.nn.Tanh,
        "sigmoid": torch.nn.Sigmoid,
        "softsign": torch.nn.Softsign,
    }

    if activation_func not in valid_activation:
        print("Unsupported Hypernetwork format, if you report it I might implement it.", path, " ", activation_func, is_layer_norm, use_dropout, activate_output, last_layer_dropout)
        return None

    out = {}

    for d in sd:
        try:
            dim = int(d)
        except:
            continue

        output = []
        for index in [0, 1]:
            attn_weights = sd[dim][index]
            keys = attn_weights.keys()

            linears = filter(lambda a: a.endswith(".weight"), keys)
            linears = list(map(lambda a: a[:-len(".weight")], linears))
            layers = []

            for i in range(len(linears)):
                lin_name = linears[i]
                last_layer = (i == (len(linears) - 1))
                penultimate_layer = (i == (len(linears) - 2))

                lin_weight = attn_weights['{}.weight'.format(lin_name)]
                lin_bias = attn_weights['{}.bias'.format(lin_name)]
                layer = torch.nn.Linear(lin_weight.shape[1], lin_weight.shape[0])
                layer.load_state_dict({"weight": lin_weight, "bias": lin_bias})
                layers.append(layer)
                if activation_func != "linear":
                    if (not last_layer) or (activate_output):
                        layers.append(valid_activation[activation_func]())
                if is_layer_norm:
                    layers.append(torch.nn.LayerNorm(lin_weight.shape[0]))
                if use_dropout:
                    if (not last_layer) and (not penultimate_layer or last_layer_dropout):
                        layers.append(torch.nn.Dropout(p=0.3))

            output.append(torch.nn.Sequential(*layers))
        out[dim] = torch.nn.ModuleList(output)

    class hypernetwork_patch:
        def __init__(self, hypernet, strength):
            self.hypernet = hypernet
            self.strength = strength
        def __call__(self, current_index, q, k, v):
            dim = k.shape[-1]
            if dim in self.hypernet:
                hn = self.hypernet[dim]
                k = k + hn[0](k) * self.strength
                v = v + hn[1](v) * self.strength

            return q, k, v

        def to(self, device):
            for d in self.hypernet.keys():
                self.hypernet[d] = self.hypernet[d].to(device)
            return self

    return hypernetwork_patch(out, strength)

class HypernetworkLoader:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": { "model": ("MODEL",),
                              "hypernetwork_name": (folder_paths.get_filename_list("hypernetworks"), ),
                              "strength": ("FLOAT", {"default": 1.0, "min": -10.0, "max": 10.0, "step": 0.01}),
                              }}
    RETURN_TYPES = ("MODEL",)
    FUNCTION = "load_hypernetwork"

    CATEGORY = "loaders"

    def load_hypernetwork(self, model, hypernetwork_name, strength):
        hypernetwork_path = folder_paths.get_full_path("hypernetworks", hypernetwork_name)
        model_hypernetwork = model.clone()
        patch = load_hypernetwork_patch(hypernetwork_path, strength)
        if patch is not None:
            model_hypernetwork.set_model_attn1_patch(patch)
            model_hypernetwork.set_model_attn2_patch(patch)
        return (model_hypernetwork,)

NODE_CLASS_MAPPINGS = {
    "HypernetworkLoader": HypernetworkLoader
}



import pytest

import hordelib
import hordelib.initialisation
from hordelib.settings import UserSettings


def test_worker_settings_singleton():
    a = UserSettings._instance
    assert a is not None
    b = UserSettings()
    assert b._instance is not None
    assert a is b._instance


def test_worker_settings_percent_check():
    assert UserSettings._is_percentage("50%") == 50
    assert UserSettings._is_percentage("50") is False
    assert UserSettings._is_percentage("%50") is False


# def test_worker_settings_no_init():
#     # _get_total_vram_mb returns 0 if hordelib is not initialised
#     # We fake this condition with _is_initialised = False
#     hordelib.initialisation._is_initialised = False
#     assert UserSettings.get_vram_to_leave_free_mb() == 0


class TestWorkerSettingsWithInit:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        hordelib.initialise()
        yield
        UserSettings._instance = None
        UserSettings._ram_to_leave_free_mb = None
        UserSettings._vram_to_leave_free_mb = None

    def test_worker_settings_properties_comparable(self):
        assert UserSettings.get_ram_to_leave_free_mb() > 0
        assert UserSettings.get_vram_to_leave_free_mb() > 0

    def test_worker_settings_set_get_ram(self):
        UserSettings.set_ram_to_leave_free_mb("50%")
        assert UserSettings.get_ram_to_leave_free_mb() > 0

    def test_worker_settings_set_get_vram(self):
        UserSettings.set_vram_to_leave_free_mb("50%")
        assert UserSettings.get_vram_to_leave_free_mb() > 0



# test_horde.py
import pytest
from PIL import Image

from hordelib.horde import HordeLib
from hordelib.shared_model_manager import SharedModelManager


class TestHordeInference:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.horde = HordeLib()

        self.default_model_manager_args = {
            # aitemplate
            # "blip": True,
            # "clip": True,
            # "codeformer": True,
            # "compvis": True,
            "controlnet": True,
            # "diffusers": True,
            # "esrgan": True,
            # "gfpgan": True,
            # "safety_checker": True,
        }
        SharedModelManager.loadModelManagers(**self.default_model_manager_args)
        assert SharedModelManager.manager is not None
        for preproc in HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys():
            SharedModelManager.manager.controlnet.download_control_type(preproc)
        assert SharedModelManager.preloadAnnotators()
        self.image = Image.open("images/test_annotator.jpg")
        self.width, self.height = self.image.size
        yield
        del self.horde
        SharedModelManager._instance = None
        SharedModelManager.manager = None

    def test_controlnet_annotator(self):
        data = {
            "sampler_name": "k_dpmpp_2m",
            "cfg_scale": 7.5,
            "denoising_strength": 1.0,
            "seed": 123456789,
            "height": 512,
            "width": 512,
            "karras": True,
            "tiling": False,
            "hires_fix": False,
            "clip_skip": 1,
            "control_type": "",
            "image_is_control": False,
            "return_control_map": True,
            "prompt": "this is not used here",
            "ddim_steps": 25,
            "n_iter": 1,
            "model": "Deliberate",
            "source_image": Image.open("images/test_annotator.jpg"),
            "source_processing": "img2img",
        }
        assert self.horde is not None

        for preproc in HordeLib.CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys():
            if preproc == "scribble":
                # Not valid for normal image input test
                continue
            assert (
                SharedModelManager.manager.controlnet.check_control_type_available(
                    preproc,
                    "stable diffusion 1",
                )
                is True
            )
            data["control_type"] = preproc
            pil_image = self.horde.basic_inference(data)
            assert pil_image is not None
            pil_image.save(f"images/annotator_{preproc}.webp", quality=90)



# This tests running hordelib standalone, as an external caller would use it.
# Call with: python -m test.run_memory_test
# You need all the deps in whatever environment you are running this.
import os
import random
import threading
import time

import psutil
from loguru import logger
from PIL import Image

import hordelib

hordelib.initialise(setup_logging=True, clear_logs=True)

from hordelib.comfy_horde import cleanup
from hordelib.horde import HordeLib
from hordelib.settings import UserSettings
from hordelib.shared_model_manager import SharedModelManager
from hordelib.utils.gpuinfo import GPUInfo

UserSettings.disable_disk_cache.activate()


BASE_KUDOS = 10


def add_model(model_name):
    logger.warning(f"Loading model {model_name}")
    SharedModelManager.manager.load(model_name)
    model_count = len(SharedModelManager.manager.compvis.get_loaded_models_names())
    logger.warning(f"{model_count} models now loaded")


def get_base_data(width=512, height=512, karras=False, steps=50, model_name="stable_diffusion"):
    return {
        "sampler_name": "k_euler",
        "cfg_scale": 7.5,
        "denoising_strength": 1.0,
        "seed": 123456789,
        "width": width,
        "height": height,
        "karras": karras,
        "tiling": False,
        "hires_fix": False,
        "clip_skip": 1,
        "control_type": None,
        "image_is_control": False,
        "return_control_map": False,
        "prompt": "an ancient llamia monster",
        "ddim_steps": steps,
        "n_iter": 1,
        "model": model_name,
    }


def do_inference(data):
    """Do some work on the GPU"""
    ITERATIONS = 1
    horde = HordeLib()
    start_time = time.time()
    for i in range(ITERATIONS):  # do n times and average the time
        pil_image = horde.basic_inference(data)
        if not pil_image:
            logger.error("Inference is failing to generate images")
        else:
            pil_image.save(f"images/stresstest/{data['model']}.webp", quality=90)
    return round((time.time() - start_time) / ITERATIONS, 2)


def calculate_kudos_cost(base_time, job_data):
    job_time = do_inference(job_data)
    return round(BASE_KUDOS * (job_time / base_time), 2)


def main():
    horde = HordeLib()
    SharedModelManager.loadModelManagers(compvis=True, controlnet=True, esrgan=True, gfpgan=True, codeformer=True)

    add_model("stable_diffusion")

    # Do some inference to warm up
    base_time = do_inference(get_base_data())
    base_time = do_inference(get_base_data())

    # base time worth 10 kudos
    logger.info(f"Calculating time for base of {BASE_KUDOS} kudos")
    base_time = do_inference(get_base_data())

    logger.info("Calculating time for steps 10")
    base_steps_10 = calculate_kudos_cost(base_time, get_base_data(steps=10))
    logger.info("Calculating time for steps 100")
    base_steps_100 = calculate_kudos_cost(base_time, get_base_data(steps=100))

    logger.info("Calculating kudos for karras")
    base_karras = calculate_kudos_cost(base_time, get_base_data(karras=True))

    logger.info("Calculating kudos for weights")
    tmpdata = get_base_data()
    tmpdata["prompt"] = "(dog:1.5) and (cat:1.5) and (mouse:1.2) and (bird:1.6) and (owl:1.9)"
    base_weights = calculate_kudos_cost(base_time, tmpdata)
    tmpdata = get_base_data()
    tmpdata["prompt"] = "dog and cat and mouse and bird and owl"
    base_no_weights = calculate_kudos_cost(base_time, tmpdata)

    logger.info("Calculating kudos for 1024x1024")
    base_1024 = calculate_kudos_cost(base_time, get_base_data(1024, 1024))

    logger.info("Calculating kudos for 2048x2048")
    base_2048 = calculate_kudos_cost(base_time, get_base_data(2048, 2048))

    logger.info("Calculating kudos for 1024x1024")
    tmpdata = get_base_data(1024, 1024)
    tmpdata["hires_fix"] = True
    base_hires_fix = calculate_kudos_cost(base_time, tmpdata)

    # Benchmark all samplers
    samplers = {}
    for sampler in horde.SAMPLERS_MAP.keys():
        logger.info(f"Calculating kudos for sampler {sampler}")
        data = get_base_data()
        data["sampler_name"] = sampler
        kudos = calculate_kudos_cost(base_time, data)
        samplers[sampler] = kudos

    # Benchmark all controlnet types
    controltypes = {}
    for controltype in horde.CONTROLNET_IMAGE_PREPROCESSOR_MAP.keys():
        logger.info(f"Calculating kudos for controlnet {controltype}")
        data = get_base_data()
        data["control_type"] = controltype
        data["source_processing"] = "img2img"
        data["source_image"] = Image.open("images/test_db0.jpg")
        kudos = calculate_kudos_cost(base_time, data)
        controltypes[controltype] = kudos

    # Results
    logger.info(f"Base time {base_time} == 10 kudos")
    logger.info(f"10 steps: {base_steps_10}")
    logger.info(f"100 steps: {base_steps_100}")
    logger.info(f"Karras: {base_karras}")
    logger.info(f"No weights: {base_no_weights}")
    logger.info(f"Weights: {base_weights}")
    logger.info(f"Kudos 1024x1024: {base_1024}")
    logger.info(f"Hires fix 1024x1024: {base_hires_fix}")
    logger.info(f"Kudos 2048x2048: {base_2048}")

    for sampler, kudos in samplers.items():
        logger.info(f"{sampler}: {kudos}")

    for controltype, kudos in controltypes.items():
        logger.info(f"{controltype}: {kudos}")


if __name__ == "__main__":
    main()
